import json
from requests.models import Response
from authlib.jose import JsonWebToken, KeySet, RSAKey
try:
    from unittest.mock import patch, Mock
except ImportError:
    from mock import patch, Mock
from cryptography.hazmat.primitives import serialization as crypto_serialization

jwt = JsonWebToken(['RS256', 'RS384', 'RS512'])

jwk_key = RSAKey.generate_key(is_private=True)
pem_key = RSAKey.generate_key(is_private=True)

def get_public_key(key):
    """Returns public key for a RSAKey object"""
    public_key = key.get_public_key().public_bytes(
        encoding=crypto_serialization.Encoding.PEM,
        format= crypto_serialization.PublicFormat.SubjectPublicKeyInfo,
    ).decode("utf-8")
    return public_key

PEM_PUBLIC_KEY = get_public_key(pem_key)

def make_id_token(sub="username",
                  iss='http://example.com',
                  aud='you',
                  exp=999999999999,  # tests will start failing in September 33658
                  iat=13151351,
                  nbf=13151351,
                  key=jwk_key,
                  kid=None,
                  **kwargs):
    payload = dict(
            iss=iss,
            aud=aud,
            exp=exp,
            iat=iat,
            nbf=nbf,
            sub=sub,
            **kwargs
        )
    # remove keys with empty values
    clean_payload = dict((k, v) for k, v in payload.items() if v)
    if kid is None:
        kid = key.as_dict().get('kid')
    return make_jwt(clean_payload, key, kid).decode('ascii')


def make_local_token():
    return make_id_token(iss="local", key=pem_key, aud="local_aud")

def make_jwt(payload, key, kid):
    jws = jwt.encode(
        {'alg': 'RS256', 'kid': kid}, payload, key=key)
    return jws


class FakeRequests(object):
    def __init__(self):
        self.responses = {}

    def set_response(self, url, content, status_code=200):
        self.responses[url] = (status_code, json.dumps(content))

    def get(self, url, *args, **kwargs):
        wanted_response = self.responses.get(url)
        if not wanted_response:
            status_code, content = 404, ''
        else:
            status_code, content = wanted_response

        response = Response()
        response._content = content.encode('utf-8')
        response.status_code = status_code

        return response


class AuthenticationTestCaseMixin(object):
    username = 'henk'

    def patch(self, thing_to_mock, **kwargs):
        patcher = patch(thing_to_mock, **kwargs)
        patched = patcher.start()
        self.addCleanup(patcher.stop)
        return patched

    def setUp(self):
        self.responder = FakeRequests()
        self.responder.set_response("http://example.com/.well-known/openid-configuration",
                                    {"jwks_uri": "http://example.com/jwks",
                                     "issuer": "http://example.com",
                                     "userinfo_endpoint": "http://example.com/userinfo"})
        self.mock_get = self.patch('requests.get')
        self.mock_get.side_effect = self.responder.get
        keys = KeySet(keys=[jwk_key])
        self.patch(
            'oidc_auth.authentication.request',
            return_value=Mock(
                status_code=200,
                json=keys.as_json
            )
        )



import logging
import time

import requests
from authlib.jose import JsonWebKey, JsonWebToken
from authlib.jose.errors import (BadSignatureError, DecodeError,
                                 ExpiredTokenError, JoseError)
import jwt as pyjwt
from django.utils.encoding import smart_str
from django.utils.translation import gettext as _
from requests import request
from requests.exceptions import HTTPError
from rest_framework.authentication import (BaseAuthentication,
                                           get_authorization_header)
from rest_framework.exceptions import AuthenticationFailed

from .settings import api_settings
from .util import cache

logger = logging.getLogger(__name__)

jwt = JsonWebToken(['RS256', 'RS384', 'RS512'])

def get_user_none(request, id_token):
    return None

class UserInfo(dict):
    """Wrapper class to allow checks to see if the object is a JWT token"""
    pass

class BearerTokenAuthentication(BaseAuthentication):
    www_authenticate_realm = 'api'

    def authenticate(self, request):
        bearer_token = self.get_bearer_token(request)
        if bearer_token is None:
            return None

        try:
            userinfo = self.get_userinfo(bearer_token)
        except HTTPError:
            msg = _('Invalid Authorization header. Unable to verify bearer token')
            raise AuthenticationFailed(msg)

        user = api_settings.OIDC_RESOLVE_USER_FUNCTION(request, userinfo)

        return user, UserInfo(userinfo)

    def get_bearer_token(self, request):
        auth = get_authorization_header(request).split()
        auth_header_prefix = api_settings.BEARER_AUTH_HEADER_PREFIX.lower()
        if not auth or smart_str(auth[0].lower()) != auth_header_prefix:
            return None

        if len(auth) == 1:
            msg = _('Invalid Authorization header. No credentials provided')
            raise AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = _(
                'Invalid Authorization header. Credentials string should not contain spaces.')
            raise AuthenticationFailed(msg)

        return auth[1]

    @cache(ttl=api_settings.OIDC_BEARER_TOKEN_EXPIRATION_TIME)
    def get_userinfo(self, token):
        userinfo_endpoint = api_settings.USERINFO_ENDPOINT
        if not userinfo_endpoint:
            raise AuthenticationFailed(_('Invalid userinfo_endpoint URL. Did not find a URL from OpenID connect '
                                         'discovery metadata nor settings.OIDC_AUTH.USERINFO_ENDPOINT.'))

        response = requests.get(userinfo_endpoint, headers={
            'Authorization': 'Bearer {0}'.format(token.decode('ascii'))})
        response.raise_for_status()

        return response.json()


class JWTToken(dict):
    """Wrapper class to allow checks to see if the object is a JWT token"""
    pass

class JSONWebTokenAuthentication(BaseAuthentication):
    """Token based authentication using the JSON Web Token standard"""

    www_authenticate_realm = 'api'

    def claims_options(self, issuer):
        _claims_options = {
            'iss': {
                'essential': True,
                'values': [issuer]
            },
            'nbf': {
                'essential': True,
            },
            'exp': {
                'essential': True,
            },
        }
        issuer_config = self.get_issuer_config(issuer)
        issuer_options = issuer_config['claims_options']
        for key, value in issuer_options.items():
            _claims_options[key] = value
        return _claims_options

    def authenticate(self, request):
        jwt_value = self.get_jwt_value(request)
        if jwt_value is None:
            return None
        payload = self.decode_jwt(jwt_value)
        self.validate_claims(payload)

        user = api_settings.OIDC_RESOLVE_USER_FUNCTION(request, payload)

        return user, JWTToken(payload)

    def get_jwt_value(self, request):
        auth = get_authorization_header(request).split()
        auth_header_prefix = api_settings.JWT_AUTH_HEADER_PREFIX.lower()

        if not auth or smart_str(auth[0].lower()) != auth_header_prefix:
            return None

        if len(auth) == 1:
            msg = _('Invalid Authorization header. No credentials provided')
            raise AuthenticationFailed(msg)
        elif len(auth) > 2:
            msg = _(
                'Invalid Authorization header. Credentials string should not contain spaces.')
            raise AuthenticationFailed(msg)

        return auth[1]

    def jwks(self, jwks_uri):
        return JsonWebKey.import_key_set(self.jwks_data(jwks_uri))

    @cache(ttl=api_settings.JWKS_EXPIRATION_TIME)
    def jwks_data(self, jwks_uri):
        r = request("GET", jwks_uri, allow_redirects=True)
        r.raise_for_status()
        return r.json()

    def decode_jwt(self, jwt_value):
        try:
            issuer = self.get_issuer_from_raw_token(jwt_value)
            key = self.get_key_for_issuer(issuer)
            id_token = jwt.decode(
                jwt_value.decode('ascii'),
                key,
                claims_options=self.claims_options(issuer),
            )
        except (BadSignatureError, DecodeError, pyjwt.exceptions.DecodeError, ValueError):
            msg = _(
                'Invalid Authorization header. JWT Signature verification failed.')
            logger.exception(msg)
            raise AuthenticationFailed(msg)
        except AssertionError:
            msg = _(
                'Invalid Authorization header. Please provide base64 encoded ID Token'
            )
            raise AuthenticationFailed(msg)

        return id_token

    def validate_claims(self, id_token):
        try:
            id_token.validate(
                now=int(time.time()),
                leeway=api_settings.LEEWAY,
            )
        except ExpiredTokenError:
            msg = _('Invalid Authorization header. JWT has expired.')
            raise AuthenticationFailed(msg)
        except JoseError as e:
            msg = _(str(type(e)) + str(e))
            raise AuthenticationFailed(msg)

    def get_key_for_issuer(self, issuer):
        issuer_config = self.get_issuer_config(issuer)
        key = issuer_config['key']
        key_type = issuer_config['type']
        if key_type not in ['JWKS', 'PEM']:
            raise ValueError(f"{key_type} is not a valid type")
        if key_type == 'JWKS':
            key = self.jwks(key)
        elif key_type == 'PEM':
            key = bytes(key, encoding='utf-8')
        return key

    def get_issuer_from_raw_token(self, token):
        claims = self.get_claims_without_validation(token)
        if 'iss' not in claims:
            raise AuthenticationFailed("Token is missing the 'iss' claim")
        return claims['iss']

    def get_claims_without_validation(self, token):
        """Raises pyjwt.exceptions.DecodeError if token could not be decoded"""
        return pyjwt.decode(token, options={"verify_signature": False})

    def get_issuer_config(self, target_issuer):
        issuer = api_settings.JWT_ISSUERS.get(target_issuer)
        if not issuer:
            raise AuthenticationFailed("Invalid 'iss' claim")
        return issuer

    def authenticate_header(self, request):
        return 'JWT realm="{0}"'.format(self.www_authenticate_realm)



from django.conf import settings
from rest_framework.settings import APISettings

USER_SETTINGS = getattr(settings, 'OIDC_AUTH', None)

DEFAULTS = {
    # Define multiple issuers in here, each with
    # a `type`, `key` and `claims_options` value.
    # The key for each issuer in the dict will be the expected value for
    # the 'iss' claim in tokens from that issuer.
    # The Claims Options can now be defined according to this documentation:
    # ref: https://docs.authlib.org/en/latest/jose/jwt.html#jwt-payload-claims-validation
    # `type` can be "PEM" or "JWKS". If "PEM", then `key` must be a public key
    # in PEM format. if "JWKS`, then `key` must be a JWKS endpoint
    'JWT_ISSUERS': {
    },

    # Time before JWKS will be refreshed
    'JWKS_EXPIRATION_TIME': 24 * 60 * 60,

    # Function to resolve user from request and token or userinfo
    'OIDC_RESOLVE_USER_FUNCTION': 'oidc_auth.authentication.get_user_none',

    # Time before bearer token validity is verified again
    'OIDC_BEARER_TOKEN_EXPIRATION_TIME': 600,

    'JWT_AUTH_HEADER_PREFIX': 'JWT',
    'BEARER_AUTH_HEADER_PREFIX': 'Bearer',

    # The Django cache to use
    'OIDC_CACHE_NAME': 'default',
    'OIDC_CACHE_PREFIX': 'oidc_auth.',

    # URL of the OpenID Provider's UserInfo Endpoint
    'USERINFO_ENDPOINT': None,

    # Leeway gives some +- leeway to when a token is active. Given in seconds
    # This helps account for clockskew
    'LEEWAY': 5,
}

# List of settings that may be in string import notation.
IMPORT_STRINGS = (
    'OIDC_RESOLVE_USER_FUNCTION',
)

api_settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)



from oidc_auth.test import PEM_PUBLIC_KEY
SECRET_KEY = 'secret'
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:'
    }
}
REST_FRAMEWORK = {
    'UNAUTHENTICATED_USER': None,
}
ROOT_URLCONF = 'tests.test_authentication'
OIDC_AUTH = {
    'USERINFO_ENDPOINT': "http://example.com/userinfo",
    'JWT_ISSUERS': {
        'http://example.com': {
            'type': "JWKS",
            'key': 'http://example.com',
            'claims_options': {
                'aud': {
                    'values': ['you'],
                    'essential': True,
                },
            }
        },
        'local': {
            'type': "PEM",
            'key': PEM_PUBLIC_KEY,
            'claims_options': {
                'aud': {
                    'values': ['local_aud'],
                    'essential': True,
                },
            }
        }
    }
}



from random import random
from unittest import TestCase

from oidc_auth.settings import api_settings
from oidc_auth.util import cache

try:
    from unittest.mock import patch, Mock, ANY
except ImportError:
    from mock import patch, Mock, ANY


class TestCacheDecorator(TestCase):
    @cache(1)
    def mymethod(self, *args):
        return random()

    @cache(1)
    def failing(self):
        raise RuntimeError()

    @cache(0)
    def notcached(self):
        return random()

    @cache(1)
    def return_none(self):
        return None

    def test_that_result_of_method_is_memoized(self):
        x = self.mymethod('a')
        y = self.mymethod('b')
        self.assertEqual(x, self.mymethod('a'))
        self.assertEqual(y, self.mymethod('b'))
        self.assertNotEqual(x, y)

    def test_that_exceptions_are_raised(self):
        with self.assertRaises(RuntimeError):
            self.failing()

    def test_that_cache_is_disabled_with_low_ttl(self):
        x = self.notcached()
        # This will fail sometimes when the RNG returns two equal numbers...
        self.assertNotEqual(x, self.notcached())

    def test_that_cache_can_store_None(self):
        self.assertIsNone(self.return_none())
        self.assertIsNone(self.return_none())

    @patch('oidc_auth.util.caches')
    def test_uses_django_cache_uncached(self, caches):
        caches['default'].get.return_value = None
        self.mymethod()
        caches['default'].get.assert_called_with('oidc_auth.mymethod', version=1)
        caches['default'].set.assert_called_with('oidc_auth.mymethod', ANY, timeout=1, version=1)

    @patch('oidc_auth.util.caches')
    def test_uses_django_cache_cached(self, caches):
        return_value = random()
        caches['default'].get.return_value = return_value
        self.assertEqual(return_value, self.mymethod())
        caches['default'].get.assert_called_with('oidc_auth.mymethod', version=1)
        self.assertFalse(caches['default'].set.called)

    @patch.object(api_settings, 'OIDC_CACHE_NAME', 'other')
    def test_respects_cache_name(self):
        caches = {
            'default': Mock(),
            'other': Mock(),
        }
        with patch('oidc_auth.util.caches', caches):
            self.mymethod()
            self.assertTrue(caches['other'].get.called)
            self.assertFalse(caches['default'].get.called)

    @patch.object(api_settings, 'OIDC_CACHE_PREFIX', 'some-other-prefix.')
    @patch('oidc_auth.util.caches')
    def test_respects_cache_prefix(self, caches):
        caches['default'].get.return_value = None
        self.mymethod()
        caches['default'].get.assert_called_once_with('some-other-prefix.mymethod', version=1)
        caches['default'].set.assert_called_once_with(
            'some-other-prefix.mymethod',
            ANY,
            timeout=1,
            version=1
        )



import sys
from datetime import datetime, timedelta

from authlib.jose.errors import BadSignatureError, DecodeError
from django.http import HttpResponse
from django.test import TestCase
from django.urls import re_path as url
from oidc_auth.authentication import (BearerTokenAuthentication,
                                      JSONWebTokenAuthentication,
                                      JWTToken,
                                      UserInfo,)
from oidc_auth.test import AuthenticationTestCaseMixin, make_id_token, make_local_token
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.permissions import BasePermission
from rest_framework.views import APIView

if sys.version_info > (3,):
    long = int
else:
    class ConnectionError(OSError):
        pass

try:
    from unittest.mock import Mock, PropertyMock, patch
except ImportError:
    from mock import Mock, PropertyMock, patch


class TokenPermission(BasePermission):
    """Checks if the token has correct permissions"""

    def has_permission(self, request, _view):
        token = request.auth  # type: JWTToken
        if not token:
            return False
        if not isinstance(token, JWTToken) and not isinstance(token, UserInfo):
            return False
        return True

class MockView(APIView):
    permission_classes = (TokenPermission,)
    authentication_classes = (
        JSONWebTokenAuthentication,
        BearerTokenAuthentication
    )

    def get(self, request):
        return HttpResponse('a')


urlpatterns = [
    url(r'^test/$', MockView.as_view(), name="testview")
]


class TestBearerAuthentication(AuthenticationTestCaseMixin, TestCase):
    urls = __name__

    def setUp(self):
        super(TestBearerAuthentication, self).setUp()
        self.openid_configuration = {
            'issuer': 'http://accounts.example.com/dex',
            'authorization_endpoint': 'http://accounts.example.com/dex/auth',
            'token_endpoint': 'http://accounts.example.com/dex/token',
            'jwks_uri': 'http://accounts.example.com/dex/keys',
            'response_types_supported': ['code'],
            'subject_types_supported': ['public'],
            'id_token_signing_alg_values_supported': ['RS256'],
            'scopes_supported': ['openid', 'email', 'groups', 'profile', 'offline_access'],
            'token_endpoint_auth_methods_supported': ['client_secret_basic'],
            'claims_supported': [
                'aud', 'email', 'email_verified', 'exp', 'iat', 'iss', 'locale',
                'name', 'sub'
            ],
            'userinfo_endpoint': 'http://sellers.example.com/v1/sellers/'
        }

    def test_using_valid_bearer_token(self):
        self.responder.set_response(
            'http://example.com/userinfo', {'sub': self.username})
        auth = 'Bearer abcdefg'
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.content.decode(), 'a', resp.content)
        self.assertEqual(resp.status_code, 200, resp.content)
        self.mock_get.assert_called_with(
            'http://example.com/userinfo', headers={'Authorization': auth})

    def test_cache_of_valid_bearer_token(self):
        self.responder.set_response(
            'http://example.com/userinfo', {'sub': self.username})
        auth = 'Bearer egergerg'
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 200, resp.content)

        # Token expires, but validity is cached
        self.responder.set_response('http://example.com/userinfo', "", 401)
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 200, resp.content)

    def test_using_invalid_bearer_token(self):
        self.responder.set_response('http://example.com/userinfo', "", 401)
        auth = 'Bearer hjikasdf'
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_cache_of_invalid_bearer_token(self):
        self.responder.set_response('http://example.com/userinfo', "", 401)
        auth = 'Bearer feegrgeregreg'
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

        # Token becomes valid
        self.responder.set_response(
            'http://example.com/userinfo', {'sub': self.username})
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 200, resp.content)

    def test_using_malformed_bearer_token(self):
        auth = 'Bearer abc def'
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_using_missing_bearer_token(self):
        auth = 'Bearer'
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_using_inaccessible_userinfo_endpoint(self):
        self.mock_get.side_effect = ConnectionError
        auth = 'Bearer'
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_get_user_info_endpoint(self):
        authentication = BearerTokenAuthentication()
        response_mock = Mock(return_value=Mock(status_code=200,
                                                json=Mock(return_value={}),
                                                raise_for_status=Mock(return_value=None)))
        with patch('oidc_auth.authentication.requests.get', response_mock):
            result = authentication.get_userinfo(b'token')
            assert result == {}


class TestJWTAuthentication(AuthenticationTestCaseMixin, TestCase):
    urls = __name__

    def test_using_valid_jwt(self):
        auth = 'JWT ' + make_id_token()
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 200, resp.content)
        self.assertEqual(resp.content.decode(), 'a', resp.content)

    def test_without_jwt(self):
        resp = self.client.get('/test/')
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_with_invalid_jwt(self):
        auth = 'JWT e30='
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_with_invalid_auth_header(self):
        auth = 'Bearer 12345'
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_with_expired_jwt(self):
        auth = 'JWT ' + make_id_token(exp=13151351)
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_with_invalid_issuer(self):
        auth = 'JWT ' + \
               make_id_token(iss='http://something.com')
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_with_invalid_audience(self):
        auth = 'JWT ' + make_id_token(aud='somebody')
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_with_too_new_jwt(self):
        auth = 'JWT ' + make_id_token(nbf=999999999999)
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_with_multiple_audiences(self):
        auth = 'JWT ' + make_id_token(aud=['you', 'me'])
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 200, resp.content)

    def test_with_invalid_multiple_audiences(self):
        auth = 'JWT ' + make_id_token(aud=['we', 'me'])
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_with_multiple_audiences_and_authorized_party(self):
        auth = 'JWT ' + \
               make_id_token(aud=['you', 'me'], azp='you')
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 200, resp.content)

    def test_with_invalid_signature(self):
        auth = 'JWT ' + make_id_token()
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth + 'x')
        self.assertEqual(resp.status_code, 401, resp.content)

    @patch('oidc_auth.authentication.jwt.decode')
    @patch('oidc_auth.authentication.logger')
    def test_decode_jwt_logs_exception_message_when_decode_throws_exception(
        self,
        logger_mock, decode
    ):
        auth = 'JWT ' + make_id_token()
        decode.side_effect = DecodeError, BadSignatureError

        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)

        self.assertEqual(resp.status_code, 401, resp.content)
        logger_mock.exception.assert_called_once_with(
            'Invalid Authorization header. JWT Signature verification failed.')

    def test_should_fail_without_aud_claim(self):
        auth = 'JWT ' + make_id_token(aud=None)
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_should_fail_without_iss_claim(self):
        auth = 'JWT ' + make_id_token(iss=None)
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_should_fail_without_exp_claim(self):
        auth = 'JWT ' + make_id_token(exp=None)
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_should_fail_without_nbf_claim(self):
        auth = 'JWT ' + make_id_token(nbf=None)
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_using_valid_jwt_and_local_issuer(self):
        auth = 'JWT ' + make_local_token()
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 200, resp.content)
        self.assertEqual(resp.content.decode(), 'a')

    def test_should_not_fail_without_sub_claim(self):
        auth = 'JWT ' + make_id_token(sub=None)
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 200, resp.content)

    def test_invalid_kid_should_give_401(self):
        auth = 'JWT ' + make_id_token(kid="invalid kid")
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 401, resp.content)

    def test_token_outside_nbf_but_within_leeway_should_be_valid(self):
        # default leeway is 5 seconds but we use 4 since there is small window
        # between token being created and it being validated
        future = datetime.now() + timedelta(seconds=4)
        auth = 'JWT ' + make_id_token(nbf=future.timestamp())
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 200, resp.content)


    def test_token_outside_exp_but_within_leeway_should_be_valid(self):
        # default leeway is 5 seconds but we use 4 since there is small window
        # between token being created and it being validated
        past = datetime.now() - timedelta(seconds=4)
        auth = 'JWT ' + make_id_token(exp=past.timestamp())
        resp = self.client.get('/test/', HTTP_AUTHORIZATION=auth)
        self.assertEqual(resp.status_code, 200, resp.content)



import functools

from django.core.cache import caches

from .settings import api_settings


class cache(object):
    """ Cache decorator that memoizes the return value of a method for some time.

    Increment the cache_version everytime your method's implementation changes 
    in such a way that it returns values that are not backwards compatible.
    For more information, see the Django cache documentation:
    https://docs.djangoproject.com/en/2.2/topics/cache/#cache-versioning
    """

    def __init__(self, ttl, cache_version=1):
        self.ttl = ttl
        self.cache_version = cache_version

    def __call__(self, fn):
        @functools.wraps(fn)
        def wrapped(this, *args):
            cache = caches[api_settings.OIDC_CACHE_NAME]
            key = api_settings.OIDC_CACHE_PREFIX + '.'.join([fn.__name__] + list(map(str, args)))
            cached_value = cache.get(key, version=self.cache_version)
            if not cached_value:
                cached_value = fn(this, *args)
                cache.set(key, cached_value, timeout=self.ttl, version=self.cache_version)
            return cached_value

        return wrapped
